# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2025 51 Degrees Mobile Experts Limited, Davidson House,
# Forbury Square, Reading, Berkshire, United Kingdom RG1 3EU.
#
# This Original Work is licensed under the European Union Public Licence
# (EUPL) v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# *********************************************************************

import string
import unittest
import os
import time
import csv
import platform

from fiftyone_devicedetection_onpremise.devicedetection_onpremise_pipelinebuilder import DeviceDetectionOnPremisePipelineBuilder

from fiftyone_devicedetection_onpremise.devicedetection_onpremise import DeviceDetectionOnPremise

from fiftyone_devicedetection_onpremise.devicedetection_datafile import DeviceDetectionDataFile

from fiftyone_pipeline_core.pipelinebuilder import PipelineBuilder

data_file = "src/fiftyone_devicedetection_onpremise/cxx/device-detection-data/51Degrees-LiteV4.1.hash"

mobile_ua = "Mozilla/5.0 (iPhone; CPU iPhone OS 11_2 like Mac OS X) AppleWebKit/604.4.7 (KHTML, like Gecko) Mobile/15C114"

# User agents test percentages
minMobilePercentage = 0.72
maxUnknownPercentage = 0.002

# User agents file

with open('src/fiftyone_devicedetection_onpremise/cxx/device-detection-data/20000 User Agents.csv', newline='') as file:
    reader = csv.reader(file)
    user_agents = list(reader)
    no_user_agents = len(user_agents)


class DeviceDetectionTests(unittest.TestCase):

    def test_pipeline_builder_shareusage_init(self):
        """!
        Tests whether the device detection pipeline builder adds the usage sharing engine when initialised with a datafile
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, licence_keys="").build()

        self.assertTrue(pipeline.flow_elements[0].datakey == "shareusage")

    def test_on_premise_engine_datafile(self):
        """!
        Tests whether a datafile (for the update service) is added when auto_update is set 
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="test", auto_update=True).build()

        self.assertTrue(isinstance(
            pipeline.flow_elements[0].data_file, DeviceDetectionDataFile))

    def test_properties_onpremise(self):
        """!
        Tests whether a properties list is created on the on
        premise engine
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        properties = pipeline.flow_elements[0].get_properties()

        self.assertTrue(len(properties.keys()) > 0)

    def test_evidencekey_filter_onpremise(self):
        """!
        Tests whether the on premise evidence key filter works
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        evidence_key_filter = pipeline.flow_elements[0].get_evidence_key_filter(
        )

        self.assertFalse(evidence_key_filter.filter_evidence_key("test.test"))
        self.assertTrue(
            evidence_key_filter.filter_evidence_key("header.user-agent"))

    def test_basic_get_onpremise(self):
        """!
        Check property lookup works
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.evidence.add("header.user-agent", mobile_ua)

        fd.process()

        self.assertTrue(fd.device.ismobile.value())

    def test_process_with_no_evidence(self):
        """!
        Process a FlowData which does not have any evidence. This should not throw an error and all 51Degrees engines should set the default aspect properties
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.process()

        self.assertTrue(fd.device.difference.has_value())

    def test_process_with_empty_user_agent(self):
        """!
        Process a FlowData with an empty user agent
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.evidence.add("header.user-agent", "")

        fd.process()

        self.assertFalse(fd.device.ismobile.has_value())

    def test_process_with_no_headers(self):
        """!
        Process a FlowData with no headers
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.evidence.add("irrelevant.evidence", "some-evidence")

        fd.process()

        self.assertFalse(fd.device.ismobile.has_value())

    def test_process_with_no_useful_headers(self):
        """!
        Process a FlowData with no useful headers
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.evidence.add("header.irrelevant", "some-evidence")

        fd.process()

        self.assertFalse(fd.device.ismobile.has_value())

    def test_case_insensitive_evidence_keys(self):
        """!
        Process a FlowData with case insensitive evidence keys
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.evidence.add("header.User-Agent", mobile_ua)

        fd.process()

        self.assertTrue(fd.device.ismobile.value())

    def test_missing_property_service_not_found_anywhere(self):
        """!
        Trigger the missing property service by requesting a property
        not available in any datafile
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.evidence.add("header.User-Agent", mobile_ua)

        fd.process()

        result = ""

        try:
            fd.device.get("notpresent")
        except Exception as e:
            result = str(e)

        self.assertEqual(
            result, "Property notpresent not found in data for element device. Please check that the element and property names are correct.")

    def test_missing_property_service_not_found_in_current(self):
        """!
        Trigger the missing property service by requesting a property
        not available in the current datafile
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.evidence.add("header.User-Agent", mobile_ua)

        # Add mock property not in datafile

        pipeline.get_element("device").properties["mock"] = {
            "datafiles": ["Enterprise"]
        }

        fd.process()

        result = ""

        try:
            fd.device.get("mock")
        except Exception as e:
            result = str(e)

        self.assertEqual(
            result, "Property mock not found in data for element device. This is because your datafile does not contain the property. The property is available in['Enterprise']")

    def test_excluded_property(self):
        """!
        Test error if accessing property that has been restricted
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file,
            licence_keys="",
            restricted_properties=["ismobile"]).build()

        fd = pipeline.create_flowdata()

        fd.evidence.add("header.User-Agent", mobile_ua)

        fd.process()

        result = ""

        try:
            fd.device.get("screenpixelsheight")
        except Exception as e:
            result = str(e)

        self.assertEqual(
            result, "Property screenpixelsheight was excluded from device")

    def test_evidencekey_filter_contains_user_agent(self):
        """!
        Test if evidence key filter with header.user-agent
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        evidence_key_filter = pipeline.flow_elements[0].get_evidence_key_filter(
        )

        self.assertTrue(
            evidence_key_filter.filter_evidence_key("header.user-agent"))

    def test_evidencekey_filter_contains_device_stock_ua(self):
        """!
        Test if evidence key filter with header.device-stock-ua
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        evidence_key_filter = pipeline.flow_elements[0].get_evidence_key_filter(
        )

        self.assertTrue(evidence_key_filter.filter_evidence_key(
            "header.device-stock-ua"))

    def test_evidencekey_filter_contains_query_params(self):
        """!
        Test if evidence key filter with query.user-agent query.device-stock-ua
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        evidence_key_filter = pipeline.flow_elements[0].get_evidence_key_filter(
        )

        self.assertTrue(evidence_key_filter.filter_evidence_key(
            "query.device-stock-ua"))
        self.assertTrue(
            evidence_key_filter.filter_evidence_key("query.user-agent"))

    def test_evidencekey_filter_case_insensitive_keys(self):
        """!
        Test evidence key filter with case insensitive keys
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        evidence_key_filter = pipeline.flow_elements[0].get_evidence_key_filter(
        )

        self.assertTrue(
            evidence_key_filter.filter_evidence_key("header.User-Agent"))
        self.assertTrue(
            evidence_key_filter.filter_evidence_key("header.user-agent"))
        self.assertTrue(
            evidence_key_filter.filter_evidence_key("header.USER-AGENT"))
        self.assertTrue(
            evidence_key_filter.filter_evidence_key("HEADER.USER-AGENT"))

    def test_evidencekey_filter_overrides(self):
        """!
        Test evidence key filter with overrides in cookies and query string
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        evidence_key_filter = pipeline.flow_elements[0].get_evidence_key_filter(
        )

        self.assertTrue(evidence_key_filter.filter_evidence_key(
            "query.51d_profileids"))
        self.assertTrue(evidence_key_filter.filter_evidence_key(
            "cookie.51d_profileids"))

    def test_profile_overrides(self):
        """!
        Test profile overrides
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.evidence.add("query.header.user-agent", "Some user agent")

        fd.evidence.add("query.51D_ProfileIds", "12280|17779|17470|18092")

        fd.process()

        self.assertEqual(fd.device.deviceid.value(), "12280-17779-17470-18092")

    def test_profile_overrides_no_headers(self):
        """!
        Test profile overrides no headers
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.evidence.add("query.51D_ProfileIds", "12280|17779|17470|18092")

        fd.process()

        self.assertEqual(fd.device.deviceid.value(), "12280-17779-17470-18092")

    def test_profile_overrides_deviceid(self):
        """!
        Test profile overrides with device ids
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.evidence.add("query.51D_ProfileIds", "12280-17779-17470-18092")

        fd.process()

        self.assertEqual(fd.device.deviceid.value(), "12280-17779-17470-18092")

    def test_device_id(self):
        """!
        Test profile overrides with device ids
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.evidence.add("header.User-Agent", mobile_ua)

        fd.process()

        self.assertEqual(fd.device.deviceid.value(), "12280-81243-82102-18092")

    def test_build_from_config(self):

        config = {
            "PipelineOptions": {
                "Elements": [
                    {
                        "elementName": "DeviceDetectionOnPremise",
                        "elementPath": "fiftyone_devicedetection_onpremise.devicedetection_onpremise",
                        "elementParameters": {
                            "data_file_path": data_file,
                            "licence_keys": ""
                        }
                    }
                ]
            }
        }

        pipeline = PipelineBuilder().build_from_configuration(config)

        fd = pipeline.create_flowdata()

        fd.evidence.add("header.User-Agent", mobile_ua)

        fd.process()

        self.assertTrue(fd.device.ismobile.value())

    def test_no_element_exists(self):
        """!
        Access flow element that doesn't exist in pipeline
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.process()

        message = ""

        try:
            message = fd.devisce
        except Exception as e:
            message = str(e)

        self.assertEqual(
            message, "There is no element data for devisce against this flow data. Available element data keys are: ['device']")

    def test_has_value_false(self):
        """!
        Test aspect property value returns false for has value
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.evidence.add("header.user-agent", "?£$%^&*")

        fd.process()

        self.assertFalse(fd.device.ismobile.has_value())

    def test_matched_user_agents(self):
        """!
        Test aspect property value returns false for has value
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.evidence.add("header.user-agent", mobile_ua)

        fd.process()

        self.assertIsInstance(fd.device.useragents.value(), list)
        self.assertEqual(len(fd.device.useragents.value()), 1)

    def test_value_types(self):
        """!
        Test type is returned correctly for property
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.evidence.add("header.user-agent", mobile_ua)

        fd.process()

        ismobile = fd.pipeline.get_element(
            "device").get_properties()["ismobile"]

        self.assertEqual(ismobile["type"], type(
            fd.device.ismobile.value()).__name__)

    def test_available_properties(self):
        """!
        Test properties that come back from getProperties actually exist in engine
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.evidence.add("header.user-agent", mobile_ua)

        fd.process()

        properties = fd.pipeline.get_element("device").get_properties()

        for engine_property in properties:
            self.assertNotEqual(fd.device.get(engine_property), None)

    def test_engine_init_performance(self):
        """!
        Test how long it takes for the engine to be initialised
        by looking at the metadata dictionary created on init
        """

        start = time.time()

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.evidence.add("header.user-agent", mobile_ua)

        fd.process()

        properties = fd.pipeline.get_element("device").get_properties()

        for engine_property in properties:
            self.assertNotEqual(fd.device.get(engine_property), None)

        end = time.time()

        total = end - start

        self.assertLess(total, 2 if platform.system() != "Darwin" else 30)

    def test_validate_data_true(self):
        """!
        Validate whether has_value returns correctly with valid evidence
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.evidence.add("header.user-agent", mobile_ua)

        fd.process()

        properties = fd.pipeline.get_element("device").get_properties()

        for engine_property in properties:
            data_property = fd.device.get(engine_property)
            if properties[engine_property]["category"] == "Device metrics":
                self.assertEqual(data_property.has_value(), True)
            elif engine_property == "deviceid":
                self.assertEqual(fd.device.deviceid.value(), "0-0-0-0")
            elif engine_property == "javascriptgethighentropyvalues":
                self.assertFalse(data_property.has_value())
            else:
                self.assertEqual(data_property.has_value(), True)

    def test_validate_data_false(self):
        """!
        Validate whether has_value returns correctly with invalid evidence
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.evidence.add("header.user-agent", "")

        fd.process()

        properties = fd.pipeline.get_element("device").get_properties()

        for engine_property in properties:
            data_property = fd.device.get(engine_property)
            if properties[engine_property]["category"] == "Device metrics":
                self.assertEqual(data_property.has_value(), True)
            elif engine_property == "deviceid":
                self.assertEqual(fd.device.deviceid.value(), "0-0-0-0")
            else:
                self.assertEqual(data_property.has_value(), False)

    def test_engine_reload(self):
        """!
        Refresh the engine
        """

        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file, usage_sharing=False, licence_keys="").build()

        fd = pipeline.create_flowdata()

        fd.process()

        fd.pipeline.get_element("device").engine.refreshData()

    def process_user_agent_list(self, pipeline):

        results = {"mobile": 0, "not_mobile": 0, "unknown": 0}

        for user_agent in user_agents:
            user_agent = user_agent[0]
            flowdata = pipeline.create_flowdata()
            flowdata.evidence.add("header.user-agent", user_agent)
            flowdata.process()

            if flowdata.device.ismobile.has_value():
                if(flowdata.device.ismobile.value()):
                    results["mobile"] += 1
                else:
                    results["not_mobile"] += 1
            else:
                results["unknown"] += 1

        return results



# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2025 51 Degrees Mobile Experts Limited, Davidson House,
# Forbury Square, Reading, Berkshire, United Kingdom RG1 3EU.
#
# This Original Work is licensed under the European Union Public Licence
# (EUPL) v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# *********************************************************************

from fiftyone_pipeline_engines.aspectdata import AspectData
from fiftyone_pipeline_core.aspectproperty_value import AspectPropertyValue

from fiftyone_pipeline_engines.missingproperty_service import MissingPropertyService

class OnPremiseMissingPropertyService(MissingPropertyService):

    def check(self, key, flow_element):

        if key in flow_element.properties:
            datafiles = flow_element.properties[key]["datafiles"]
            if(not flow_element.engine.getProduct() in datafiles):
                raise Exception('Property ' + key + ' not found in data for element ' + flow_element.datakey + '. This is because your datafile does not contain the property. The property is available in' + str(datafiles))
        else:
            raise Exception('Property ' + key + ' not found in data for element ' + flow_element.datakey + '. Please check that the element and property names are correct.')

class SwigData(AspectData):

    """!
    Extention of AspectData made for retrieving results
    created by the SWIG wrapper of the On Premise Device Detection engine
    """

    def __init__(self, flow_element, swig_results):
        
        """!
        Constructor for swig_data
        @type flow_element: FlowElement
        @param flow_element: FlowElement / Engine the data is keyed under
        @type swig_results: ResultsHashSwig
        @param swig_results: Results from the SWIG wrapper's process method
        """

        super(SwigData, self).__init__(flow_element)

        self.swig_results = swig_results

        self.missing_property_service = OnPremiseMissingPropertyService()

    def get_internal(self, key):

        """!
        Get property values out of the SWIG wrapper
        @type key: string
        @param key: The key of the property to retrieve
        @returns mixed
        """

        #   start with special properties

        # Lowercase key

        key = key.lower()

        if key == "deviceid":
            return AspectPropertyValue(value = self.swig_results.getDeviceId())

        if key == 'useragents':
            useragents = []
            for i in range(self.swig_results.getUserAgents()):
                useragents.append(self.swig_results.getUserAgent(i))
            return AspectPropertyValue(value = useragents)

        if key == 'difference':
            return AspectPropertyValue(value = self.swig_results.getDifference())

        if key == 'method':
            return AspectPropertyValue(value = self.swig_results.getMethod())

        if key == 'matchednodes':
            return AspectPropertyValue(value = self.swig_results.getMatchedNodes())

        if key == 'drift':
            return AspectPropertyValue(value = self.swig_results.getDrift())

        if key == 'iterations':
            return AspectPropertyValue(value = self.swig_results.getIterations())
            
        # End special properties

        this_property = self.flow_element.properties[key]

        if this_property:

            value = None

            if this_property["type"] == "bool":
                value = self.swig_results.getValueAsBool(this_property["name"])

            if this_property["type"] == "string":
                value = self.swig_results.getValueAsString(this_property["name"])

            if this_property["type"] == "javascript":
                value = self.swig_results.getValueAsString(this_property["name"])
            
            if this_property["type"] == "int":
                value = self.swig_results.getValueAsInteger(this_property["name"])

            if this_property["type"] == "double":
                value = self.swig_results.getValueAsDouble(this_property["name"])

            if this_property["type"] == "string[]":
                value = self.swig_results.getValues(this_property["name"])
   

            # Check for value / no value message

            if value.hasValue():
                
                # Convert value to list if needed
                if this_property["type"] == 'string[]':
                    return AspectPropertyValue(value=list(value.getValue()))

                return AspectPropertyValue(value=value.getValue())
            
            else:

                return AspectPropertyValue(no_value_message=value.getNoValueMessage())



# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2025 51 Degrees Mobile Experts Limited, Davidson House,
# Forbury Square, Reading, Berkshire, United Kingdom RG1 3EU.
#
# This Original Work is licensed under the European Union Public Licence
# (EUPL) v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# *********************************************************************

import os
import unittest

from fiftyone_pipeline_core.pipelinebuilder import PipelineBuilder
from fiftyone_devicedetection_onpremise.devicedetection_onpremise import DeviceDetectionOnPremise
from fiftyone_devicedetection_onpremise.swig_data import SwigData
from fiftyone_devicedetection_shared.utils import *

data_file = "./src/fiftyone_devicedetection_onpremise/cxx/device-detection-data/51Degrees-LiteV4.1.hash"

# Create a simple pipeline to access the engine with and process it with flow data
deviceDetectionOnPremiseEngine = DeviceDetectionOnPremise(
            data_file_path = data_file, 
            licence_keys = "",
            auto_update=False)
pipeline = PipelineBuilder() \
            .add(deviceDetectionOnPremiseEngine) \
            .build()

class SwigTests(unittest.TestCase):

    # Mock the native code parts so we are only testing the wrapper.
    class MockSwigResults:
        class MockSwigValue:
            def __init__(self, value, noValueMessage):
                self.value = value
                self.noValueMessage = noValueMessage

            def hasValue(self):
                return True

            def getNoValueMessage(self):
                return self.noValueMessage

            def getValue(self):
                return self.value

        def __init__(self, values):
            self.values = values

        def getAvailableProperties(self):
            return self.values.len()

        def containsProperty(self, propertyName):
            return propertyName.lower() in self.values

        def getProperties(self):
            return self.values.keys()

        def getPropertyName(self, requiredPropertyIndex):
            return self.values.keys()[requiredPropertyIndex]
            
        def getValues(self, propertyName):
            return self.MockSwigValue(self.values[propertyName.lower()], "")

        def getValueAsString(self, propertyName):
            return self.MockSwigValue(self.values[propertyName.lower()], "")
            
        def getValueAsBool(self, propertyName):
            return self.MockSwigValue(self.values[propertyName.lower()], "")
            
        def getValueAsInteger(self, propertyName):
            return self.MockSwigValue(self.values[propertyName.lower()], "")
            
        def getValueAsDouble(self, propertyName):
            return self.MockSwigValue(self.values[propertyName.lower()], "")
            
    # Test fetching a value of a certain type through the wrapper/
    def run_type_tests(self, property, type, value):
        
        engine = pipeline.flow_elements_list["device"]

        # Keep hold of the existing properties to reset at the end.
        realProperties = engine.properties

        # Set up the properties in the engine.
        engine.properties = {
            property.lower(): {"name": property, "type": type, "datafiles": [engine.engine.getProduct()]}
        }

        # Create a new SwigData with the native results mocked.
        values = {property.lower(): value}
        device = SwigData(engine, self.MockSwigResults(values))

        # Check the values are returned and that they are correct
        self.assertIsNotNone(device[property.lower()].value())
        self.assertEqual(value, device[property.lower()].value())

        # Reset the properties in the enigne.
        engine.properties = realProperties

    def test_string(self):
        self.run_type_tests("StringValue", "string", "some string")
    def test_bool(self):
        self.run_type_tests("BoolValue", "bool", True)
    def test_int(self):
        self.run_type_tests("IntValue", "int", 51)
    def test_double(self):
        self.run_type_tests("DoubleValue", "double", 5.1)
    def test_list(self):
        self.run_type_tests("ListValue", "string[]", ["some string", "another string"])



# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2025 51 Degrees Mobile Experts Limited, Davidson House,
# Forbury Square, Reading, Berkshire, United Kingdom RG1 3EU.
#
# This Original Work is licensed under the European Union Public Licence
# (EUPL) v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# *********************************************************************
import hashlib
import os
import threading
import time
import unittest
import gzip
import shutil
import requests
from fiftyone_pipeline_core.pipelinebuilder import PipelineBuilder
from fiftyone_pipeline_engines.datafile_update_service import DataFileUpdateService
from flask import Flask, send_from_directory, make_response, request

from fiftyone_devicedetection_examples.example_utils import ExampleUtils
from fiftyone_devicedetection_onpremise.devicedetection_datafile import DeviceDetectionDataFile
from fiftyone_devicedetection_onpremise.devicedetection_onpremise_pipelinebuilder import \
    DeviceDetectionOnPremisePipelineBuilder

data_file = ExampleUtils.find_file("51Degrees-LiteV4.1.hash")
hash_file_path = data_file
fixtures_path="tests/fixtures"
archive_file_path = f"{fixtures_path}/51Degrees-LiteV4.1.gz"


class DeviceDetectionTests(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.port = 5689
        cls._run_web_server(
            cls.port,
            cls._init_hash_file_archive()
        )

        time.sleep(2)

    @classmethod
    def tearDownClass(cls) -> None:
        requests.get(f"http://127.0.0.1:{cls.port}/shutdown")
        cls.clear()

    def test_on_premise_data_update_url(self):
        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file,
            usage_sharing=False,
            licence_keys="test",
            update_on_start=True,
            data_update_verify_md5=False,
            data_update_url=f"http://127.0.0.1:{self.port}"
        ).build()

        pipeline.create_flowdata().process()

        flow_element_data_file = pipeline.flow_elements[0].data_file
        self.assertIsInstance(flow_element_data_file, DeviceDetectionDataFile)
        self.assertFalse(flow_element_data_file.verify_md5)
        self._assertSignedUrl(flow_element_data_file.get_update_url())

    def test_on_premise_data_old_property_data_file_update_base_url(self):
        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file,
            usage_sharing=False,
            licence_keys="test",
            update_on_start=True,
            data_update_verify_md5=False,
            data_file_update_base_url=f"http://127.0.0.1:{self.port}"
        ).build()

        pipeline.create_flowdata().process()

        flow_element_data_file = pipeline.flow_elements[0].data_file
        self.assertIsInstance(flow_element_data_file, DeviceDetectionDataFile)
        self.assertFalse(flow_element_data_file.verify_md5)
        self._assertSignedUrl(flow_element_data_file.get_update_url())

    def test_on_premise_data_update_url_config(self):
        config = self._build_pipeline_config(
            {
                "data_file_path": f"{data_file}",
                "usage_sharing": False,
                "licence_keys": "test",
                "update_on_start": True,
                "data_update_verify_md5": False,
                "data_update_url": f"http://127.0.0.1:{self.port}"
            }
        )

        pipeline = PipelineBuilder().build_from_configuration(config)
        pipeline.create_flowdata().process()

        flow_element_data_file = pipeline.flow_elements[0].data_file
        self.assertIsInstance(flow_element_data_file, DeviceDetectionDataFile)
        self.assertFalse(flow_element_data_file.verify_md5)
        self._assertSignedUrl(flow_element_data_file.get_update_url())

    def test_on_premise_data_config_old_property_data_file_update_base_url(self):
        config = self._build_pipeline_config(
            {
                "data_file_path": f"{data_file}",
                "usage_sharing": False,
                "licence_keys": "test",
                "update_on_start": True,
                "data_update_verify_md5": False,
                "data_file_update_base_url": f"http://127.0.0.1:{self.port}"
            }
        )

        pipeline = PipelineBuilder().build_from_configuration(config)
        pipeline.create_flowdata().process()

        flow_element_data_file = pipeline.flow_elements[0].data_file
        self.assertIsInstance(flow_element_data_file, DeviceDetectionDataFile)
        self.assertFalse(flow_element_data_file.verify_md5)
        self._assertSignedUrl(flow_element_data_file.get_update_url())

    def test_on_premise_data_update_verify_md5(self):
        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file,
            usage_sharing=False,
            licence_keys="test",
            update_on_start=True,
            data_update_verify_md5=True,
            data_update_url=f"http://127.0.0.1:{self.port}/md5"
        ).build()

        pipeline.create_flowdata().process()

        flow_element_data_file = pipeline.flow_elements[0].data_file
        self.assertIsInstance(flow_element_data_file, DeviceDetectionDataFile)
        self.assertTrue(flow_element_data_file.verify_md5)
        self._assertSignedUrl(flow_element_data_file.get_update_url())

    def test_on_premise_data_update_verify_md5_old_property(self):
        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file,
            usage_sharing=False,
            licence_keys="test",
            update_on_start=True,
            verify_md5=True,
            data_update_url=f"http://127.0.0.1:{self.port}/md5"
        ).build()

        pipeline.create_flowdata().process()

        flow_element_data_file = pipeline.flow_elements[0].data_file
        self.assertIsInstance(flow_element_data_file, DeviceDetectionDataFile)
        self.assertTrue(flow_element_data_file.verify_md5)
        self._assertSignedUrl(flow_element_data_file.get_update_url())

    def test_on_premise_data_update_verify_md5_config(self):
        config = self._build_pipeline_config(
            {
                "data_file_path": f"{data_file}",
                "usage_sharing": False,
                "licence_keys": "test",
                "update_on_start": True,
                "data_update_verify_md5": True,
                "data_update_url": f"http://127.0.0.1:{self.port}/md5"
            }
        )

        pipeline = PipelineBuilder().build_from_configuration(config)
        pipeline.create_flowdata().process()

        flow_element_data_file = pipeline.flow_elements[0].data_file
        self.assertIsInstance(flow_element_data_file, DeviceDetectionDataFile)
        self.assertTrue(flow_element_data_file.verify_md5)
        self._assertSignedUrl(flow_element_data_file.get_update_url())

    def test_on_premise_data_update_verify_md5_config_old_property(self):
        config = self._build_pipeline_config(
            {
                "data_file_path": f"{data_file}",
                "usage_sharing": False,
                "licence_keys": "test",
                "update_on_start": True,
                "verify_md5": True,
                "data_update_url": f"http://127.0.0.1:{self.port}/md5"
            }
        )

        pipeline = PipelineBuilder().build_from_configuration(config)
        pipeline.create_flowdata().process()

        flow_element_data_file = pipeline.flow_elements[0].data_file
        self.assertIsInstance(flow_element_data_file, DeviceDetectionDataFile)
        self.assertTrue(flow_element_data_file.verify_md5)
        self._assertSignedUrl(flow_element_data_file.get_update_url())

    def test_on_premise_data_update_use_url_formatter(self):
        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file,
            usage_sharing=False,
            licence_keys="test",
            update_on_start=True,
            data_update_verify_md5=False,
            data_update_url=f"http://127.0.0.1:{self.port}/",
            data_update_use_url_formatter=False
        ).build()

        pipeline.create_flowdata().process()

        flow_element_data_file = pipeline.flow_elements[0].data_file
        self.assertIsInstance(flow_element_data_file, DeviceDetectionDataFile)
        self._assertNotSignedUrl(flow_element_data_file.get_update_url())

    def test_on_premise_data_update_use_url_formatter_config(self):
        config = self._build_pipeline_config(
            {
                "data_file_path": f"{data_file}",
                "usage_sharing": False,
                "licence_keys": "test",
                "update_on_start": True,
                "data_update_verify_md5": False,
                "data_update_url": f"http://127.0.0.1:{self.port}",
                "data_update_use_url_formatter": False
            }
        )

        pipeline = PipelineBuilder().build_from_configuration(config)
        pipeline.create_flowdata().process()

        flow_element_data_file = pipeline.flow_elements[0].data_file
        self.assertIsInstance(flow_element_data_file, DeviceDetectionDataFile)
        self._assertNotSignedUrl(flow_element_data_file.get_update_url())

    def test_on_premise_data_update_status_rate_limit(self):
        pipeline = DeviceDetectionOnPremisePipelineBuilder(
            data_file_path=data_file,
            usage_sharing=False,
            licence_keys="test",
            update_on_start=True,
            verify_md5=True,
            data_update_url=f"http://127.0.0.1:{self.port}/status/429"
        ).build()

        pipeline.create_flowdata().process()

        flow_element_data_file = pipeline.flow_elements[0].data_file
        self.assertIsInstance(flow_element_data_file, DeviceDetectionDataFile)
        self.assertTrue(flow_element_data_file.verify_md5)
        self._assertSignedUrl(flow_element_data_file.get_update_url())

    def _build_pipeline_config(self, parameters):
        return {
            "PipelineOptions": {
                "Elements": [
                    {
                        "elementName": "DeviceDetectionOnPremise",
                        "elementPath": "fiftyone_devicedetection_onpremise.devicedetection_onpremise",
                        "elementParameters": parameters
                    }
                ]
            }
        }

    def _assertSignedUrl(self, url):
        self.assertIn(f"http://127.0.0.1:{self.port}", url)
        self.assertIn("Type", url)
        self.assertIn("licenseKeys", url)

    def _assertNotSignedUrl(self, url):
        self.assertIn(f"http://127.0.0.1:{self.port}", url)
        self.assertNotIn("Type", url)
        self.assertNotIn("licenseKeys", url)


    @staticmethod
    def _run_web_server(port, file_md5):
        def run():
            app = Flask(__name__)

            @app.route("/")
            def root():
                return send_from_directory("fixtures", "51Degrees-LiteV4.1.gz", as_attachment=True)

            @app.route("/md5")
            def md5():
                response = make_response(
                    send_from_directory("fixtures", "51Degrees-LiteV4.1.gz", as_attachment=True)
                )
                response.headers["content-md5"] = file_md5

                return response

            @app.route("/status/<status>")
            def response_status(status):
                return "Record not found", status

            @app.route("/shutdown")
            def shutdown():
                from flask import request
                request.environ["werkzeug.server.shutdown"]()
                return "OK", 200

            app.run(host="127.0.0.1", port=port)

        thread = threading.Thread(target=run, daemon=True)
        thread.start()

    @staticmethod
    def _init_hash_file_archive():
        os.makedirs(fixtures_path)
        with open(hash_file_path, "rb") as f_in:
            with gzip.open(archive_file_path, "wb", compresslevel=1) as f_out:
                shutil.copyfileobj(f_in, f_out)

        return hashlib.md5(open(archive_file_path, "rb").read()).hexdigest()

    @staticmethod
    def clear():
        os.remove(archive_file_path)
        os.rmdir(fixtures_path)



# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2025 51 Degrees Mobile Experts Limited, Davidson House,
# Forbury Square, Reading, Berkshire, United Kingdom RG1 3EU.
#
# This Original Work is licensed under the European Union Public Licence
# (EUPL) v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# *********************************************************************

from __future__ import absolute_import
import multiprocessing
import os.path

from .constants import *

from fiftyone_pipeline_engines.engine import Engine
from fiftyone_pipeline_engines.aspectdata_dictionary import AspectDataDictionary

from fiftyone_pipeline_core.basiclist_evidence_keyfilter import BasicListEvidenceKeyFilter

from .DeviceDetectionHashEngineModule import *

from .swig_data import SwigData
from .devicedetection_datafile import DeviceDetectionDataFile


class DeviceDetectionOnPremise(Engine):
    """!
    The Device Detection On Premise engine uses a datafile to
    generate a list of properties about a device based on evidence
    supplied to it
    """

    def __init__(
        self,
        data_file_path=None,
        data=None,
        auto_update=None,
        cache=None,
        restricted_properties=None,
        licence_keys=None,
        download=True,
        max_matched_useragent_length=None,
        drift=None,
        difference=None,
        allow_unmatched=None,
        file_system_watcher=False,
        polling_interval=30,
        update_time_maximum_randomisation=10,
        create_temp_data_copy=True,
        update_matched_useragent=False,
        performance_profile="LowMemory",
        reuse_temp_file=False,
        concurrency=multiprocessing.cpu_count(),
        update_on_start=False,
        data_update_url="https://distributor.51degrees.com/api/v2/download",
        use_predictive_graph=None,
        use_performance_graph=None,
        data_update_use_url_formatter=True,
        data_update_verify_md5=True,
        data_update_product_type=None,
        **kwargs
    ):
        """!

        Constructor for the DeviceDetection On Premise Engine.

        @type data_file_path: string
        @param data_file_path: path to the data file
        @type data: binary
        @param data: in memory version of datafile
        @type auto_update: bool
        @param auto_update: whether to automatically update the datafile when required
        @type restricted_properties: list
        @param restricted_properties: List of property keys to limit the engine to
        @type license_keys: string
        @param licence_keys: licencekeys to use for the data file update service
        @type download: bool
        @param download: whether to download the datafile or keep it in memory when it is returned from the datafile update service
        @type max_matched_useragents_length : int
        @param Number of characters to consider in the matched User-Agent. Ignored if update_matched_useragent is false
        @type update_matched_useragent:
        @param: update_matched_useragent: True if the detection should record the matched characters from the target User-Agent
        @type drift: int
        @param drift: Set maximum drift in hash position to allow when processing HTTP headers
        @type difference: int
        @param difference: set the maximum difference to allow when processing HTTP headers. The difference is the difference in hash value between the hash that was found, and the hash that is being searched for. By default this is 0.
        @type allow_unmatched : bool
        @param allow_unmatched:  True if there should be at least one matched node in order for the results to be considered valid. By default, this is false
        @type performance_profile: string
        @param performance_profile: options are: LowMemory, MaxPerformance, Balanced, BalancedTemp, HighPerformance
        @type reuse_temp_file: bool
        @param reuse_temp_file: Indicates that an existing temp file may be used. This should be selected if multiple instances wish to use the same file to prevent high disk usage.
        @type concurrency: int
        @param concurrency: defaults to the number of cpus in the machine
        @type update_on_start : bool
        @param update_on_start : When this is set to true the datafile is updated / downloaded immediately on initialization. This is useful if no initial datafile is present.
        @type file_system_watcher: bool
        @param file_system_watcher: whether to check the datafile's path for changes and update the connected FlowElement's data
        @type polling_interval: int
        @param polling_interval: How often to poll for updates to the datafile (minutes)
        @type update_time_maximum_randomisation : int
        @param update_time_maximum_randomisation :
        Maximum randomisation offset in seconds to polling time interval
        @type data_update_verify_md5 : bool
        @type create_temp_copy: bool
        @param create_temp_copy: whether to copy datafile to temporary location when updating
        @type data_update_url: string
        @param data_update_url: base url for the datafile update service
        @type use_performance_graph: bool
        @param use_performance_graph: True if the performance optimized graph should be used for processing
        @type use_predictive_graph: bool
        @param use_predictive_graph: True if the predictive optimized graph should be used for processing
        @param data_update_use_url_formatter: This must be set to false to prevent the API from appending the query string parameters that are required when calling the 51Degrees Distributor service.
        @type data_update_use_url_formatter: bool

        """

        super(DeviceDetectionOnPremise, self).__init__()

        self.datakey = "device"

        if not data_file_path and not data:
            raise Exception("data_file_path or data is required")

        if data_file_path:
            if os.path.isfile(data_file_path) == False and not update_on_start:
                raise Exception("There is no file at " + data_file_path)

            extention = os.path.splitext(data_file_path)[1]

            if extention != ".hash":
                raise Exception("Data file name must end in .hash")
        else:
            self.data = data

        if not licence_keys and licence_keys != "":
            raise Exception("licence key is required. A key can be obtained from the 51Degrees website: https://51degrees.com/pricing. If you do not wish to use a key then you can specify an empty string, but this will cause automatic updates to be disabled.")

        #   Create SWIG wrapper vector for restricted properties and add

        restricted_properties_vector = VectorStringSwig()

        if restricted_properties:
            for restricted_property in restricted_properties:
                restricted_properties_vector.append(restricted_property)
            # Add the metric properties to the restricted properties as they
            # are always available
            for property in self.get_metric_properties():
                restricted_properties.append(property)
            self.set_restricted_properties(restricted_properties)

        properties_list = RequiredPropertiesConfigSwig(restricted_properties_vector)

        #  Assemble configuration

        config = ConfigHashSwig()

        available_performance = ["LowMemory", "MaxPerformance", "Balanced", "BalancedTemp", "HighPerformance"]

        if performance_profile not in available_performance:
            raise Exception("Not a valid performance profile")

        if performance_profile == "LowMemory":
            config.setLowMemory()
        if performance_profile == "MaxPerformance":
            config.setMaxPerformance()
        if performance_profile == "Balanced":
            config.setBalanced()
        if performance_profile == "BalancedTemp":
            config.setBalancedTemp()
        if performance_profile == "HighPerformance":
            config.setHighPerformance()

        # Don't set an http_ prefix on evidence keys

        config.setUseUpperPrefixHeaders(False)

        # Set performance graph of predictive graph if specified

        if(use_performance_graph):
            config.setUsePerformanceGraph(True)

        if(use_predictive_graph):
            config.setUsePredictiveGraph(True)

        # If auto update is enabled then we must use a temporary copy of the file

        if auto_update == True:
            config.setUseTempFile(True)

        if create_temp_data_copy == True:
            config.setUseTempFile(True)

        config.setReuseTempFile(reuse_temp_file)
        config.setUpdateMatchedUserAgent(update_matched_useragent)

        if allow_unmatched == True or allow_unmatched == False:
            config.setAllowUnmatched(allow_unmatched)

        if max_matched_useragent_length:
            config.setMaxMatchedUserAgentLength(max_matched_useragent_length)

        if drift:
            config.setDrift(drift)

        if (difference):
            config.setDifference(difference)

        config.setConcurrency(concurrency)

        # Make engine

        self.config = config
        self.properties_list = properties_list
        self.data_file_path = data_file_path

        # Disable features that require a licence key if one was not supplied.

        if licence_keys:
            auto_update = auto_update and (len(licence_keys) > 0)
            update_on_start = update_on_start and (len(licence_keys) > 0)
        else:
            auto_update = False
            update_on_start = False

        if auto_update or update_on_start or file_system_watcher:

            # Construct DataFile

            update_url_params = {
                "base_url": data_update_url
            }

            if data_update_use_url_formatter:
                update_url_params = {
                    "base_url": data_update_url,
                    "type": "HashV41",
                    "product": data_update_product_type if data_update_product_type else None,
                    "license_keys": licence_keys if licence_keys else None
                }

                # Remove keys with None values
                update_url_params = {k: v for k, v in update_url_params.items() if v is not None}

            data_file = DeviceDetectionDataFile(flow_element=self, identifier="HashV41", verify_md5=data_update_verify_md5, auto_update = auto_update, update_on_start = update_on_start, decompress= True, path = data_file_path, download=download, file_system_watcher=file_system_watcher, polling_interval=polling_interval, update_time_maximum_randomisation=update_time_maximum_randomisation, update_url_params = update_url_params, data_update_use_url_formatter = data_update_use_url_formatter)

            self.register_data_file(data_file)

        self.data_file_path = data_file_path
        self.config = config
        self.properties_list = properties_list

    def init_engine(self):

        """!

        Function for initialising the engine, wrapped like this so that an engine can be initialised once the datafile is retrieved if update_on_start is set to true. If this is the case, processing is held until the data file is downloaded and available.

        """

        if(self.data_file_path):
            data = self.data_file_path
        else:
            data = self.data

        engine = EngineHashSwig(data, self.config, self.properties_list)

        self.engine = engine

        # Get keys and add to evidenceKey filter

        evidence_keys = engine.getKeys()

        evidence_keys_list = []

        for x in range(evidence_keys.size()):
            evidence_keys_list.append(evidence_keys.__getitem__(x).lower())

        self.evidence_keys_list = evidence_keys_list

        # Get properties list

        properties_internal = engine.getMetaData().getProperties()

        properties = {}

        for x in range(properties_internal.getSize()):
            current_property = properties_internal.getByIndex(x)
            if current_property.getAvailable():
                properties[current_property.getName().lower()] = {
                "name": current_property.getName(),
                "type": current_property.getType(),
                "datafiles": list(current_property.getDataFilesWherePresent()),
                "category": current_property.getCategory(),
                "description": current_property.getDescription(),
                "component": self.get_component(current_property)
            }

        self.properties = properties

        # Special properties
        for key, property in self.get_metric_properties().items():
            self.properties[key] = property

    def get_metric_properties(self):
        return { "deviceid": {
          "name": 'DeviceId',
          "type": 'string',
          "category": 'Device metrics',
          "description": DEVICE_ID_DESCRIPTION,
          "component": "Match Metrics"
        },
        "useragents": {
          "name": 'UserAgents',
          "type": 'string',
          "category": 'Device metrics',
          "description": USER_AGENTS_DESCRIPTION,
          "component": "Match Metrics"
        },
        "difference": {
          "name": 'Difference',
          "type": 'int',
          "category": 'Device metrics',
          "description": DIFFERENCE_DESCRIPTION,
          "component": "Match Metrics"
        },
        "matchednodes": {
            "name": 'MatchedNodes',
            "type": 'int',
            "category": 'Device metrics',
            "description": MATCHED_NODES_DESCRIPTION,
            "component": "Match Metrics"
        },
        "drift": {
            "name": 'Drift',
            "type": 'int',
            "category": 'Device metrics',
            "description": DRIFT_DESCRIPTION,
            "component": "Match Metrics"
        },
        "iterations": {
            "name": 'Iterations',
            "type": 'int',
            "category": 'Device metrics',
            "description": ITERATIONS_DESCRIPTION,
            "component": "Match Metrics"
        },
        "method": {
            "name": 'Method',
            "type": 'string',
            "category": 'Device metrics',
            "description": METHOD_DESCRIPTION,
            "component": "Match Metrics"
        }}

    def get_evidence_key_filter(self):
        """!
        Returns evidence key filter for the on premise engine.
        Generated via a list of evidence keys from the swig engine

        """

        return BasicListEvidenceKeyFilter(self.evidence_keys_list)


    def on_registration(self, pipeline):

        """!

        Function called after the engine is registered with a pipeline
        the engine base class registers the datafile but we also need
        to initialise the SWIG wrapper so call self.init_engine() here
        after the datafile is registered.

        """

        super(DeviceDetectionOnPremise, self).on_registration(pipeline)
        self.init_engine()

    def process_internal(self, flow_data):

        """!

        Internal process method of the Device Detection On Premise engine
        Gets properties through the SWIG wrapper and stores them in a
        SwigData extension of the ElementData class. Each property value (or lack of) is returned in an AspectPropertyData wrapper from
        the core Pipeline library.

        """

        evidence = EvidenceDeviceDetectionSwig()

        for key, value in flow_data.evidence.get_all().items():
            evidence.__setitem__(key, value)

        result = self.engine.process(evidence)

        data = SwigData(self, result)

        flow_data.set_element_data(data)

    def get_component(self, property):
        return self.engine.getMetaData().getComponentForProperty(property).getName()



# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2025 51 Degrees Mobile Experts Limited, Davidson House,
# Forbury Square, Reading, Berkshire, United Kingdom RG1 3EU.
#
# This Original Work is licensed under the European Union Public Licence
# (EUPL) v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# *********************************************************************

"""
setup.py file for SWIG
"""
import sys
import platform
from setuptools import setup, Extension
from setuptools.command import develop
from distutils import sysconfig
from Cython.Distutils import build_ext
import os
import io

# Read a text file and return the content as a string.
def read(file_name):

    """Read a text file and return the content as a string."""
    try:
        with io.open(
            os.path.join(os.path.dirname(__file__), file_name), encoding="utf-8"
        ) as f:
            return f.read().strip()
    except:
        return "0.0.0"
        
def get_version():

    version = None
    
    # Try to get the correct version from platform.
    is_64bits = sys.maxsize > 2**32
    if is_64bits:
        if 'Windows' in platform.system():
            version = "win_64_"
        if 'Linux' in platform.system():
            version = "linux_64_"
        if 'Darwin' in platform.system():
            version = "darwin_64_"
    else:
        if 'Windows' in platform.system():
            version = "win_32_"
        if 'Linux' in platform.system():
            version = "linux_32_"
        if 'Darwin' in platform.system():
            # No 32 bit python on Mac
            version = None
            raise Exception("Sorry, No 32 bit Python on Mac")

    python_version = str(sys.version_info[0])
    return version + python_version

class NoSuffixBuilder(build_ext):
    def get_ext_filename(self, ext_name):
        filename = super().get_ext_filename(ext_name)
        suffix = sysconfig.get_config_var('EXT_SUFFIX')
        if suffix is None:
            suffix = sysconfig.get_config_var('SO')

        print("**" + filename + "**\r\n")
        print("**" + suffix + "**\r\n")
        
        # Get the version from the suffix.
        version = get_version()
 
        ext = os.path.splitext(filename)[1]

        if self.inplace:
            return "./" + filename.replace(suffix, "") + ext
        else:
            return filename.replace(suffix, "") + ext

define_macros = []
extra_compile_args = []
extra_link_args = []
# Extra compilation flags for C library
cflags = []

if sys.platform != "win32":
    extra_compile_args.extend([
        '-fPIC',
        '-std=gnu++11',
        '-Wall',
        '-Werror'
    ])
    cflags.extend([
        '-std=gnu11',
        '-Wall',
        '-Werror',
        '-Wno-strict-prototypes',
        '-Wno-unused-variable',
        '-Wno-missing-braces',
        '-Wno-strict-aliasing'
    ])


if sys.platform == "win32":
    extra_compile_args.extend([
        '/D_CRT_SECURE_NO_WARNINGS',
        '/D_UNICODE',
        '/DUNICODE',
        '/W4',
        # '/WX',
        '/wd4127',
        '/wd4456',
        '/wd4701',
        '/wd4703',
        '/wd4706'
    ])
    cflags.extend([
        '/D_CRT_SECURE_NO_WARNINGS',
        '/D_UNICODE',
        '/DUNICODE',
        '/W4',
        # '/WX'
    ])
    extra_link_args.extend([
        '/WX'
    ])

if sys.platform == "linux":
    extra_link_args.extend([
        '-latomic'
    ])

if sys.platform == 'darwin':
    # This is flaged in CPython
    extra_compile_args.extend([
        '-DGCC_ENABLE_CPP_EXCEPTIONS=YES'
    ])
    if sys.version_info[0] == 3 and sys.version_info[1] == 8:
        extra_compile_args.extend([
            '-Wno-deprecated-declarations'
    ])
    cflags.extend([
        '-Wno-atomic-alignment'
    ])

DeviceDetectionHashEngineModule = Extension('_DeviceDetectionHashEngineModule',
        sources=[
        # Python Wrapper
        # "fiftyone_devicedetection_onpremise/hash_python.i"
        "src/fiftyone_devicedetection_onpremise/hash_python_wrap.cxx"
        ],
        define_macros=define_macros,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        language='c++',
    )

cpplib = ('cpplib', {
        'sources': [
            # Common C++
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/CollectionConfig.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/ComponentMetaData.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/ConfigBase.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/Date.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/EngineBase.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/EvidenceBase.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/Exceptions.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/MetaData.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/ProfileMetaData.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/PropertyMetaData.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/RequiredPropertiesConfig.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/ResultsBase.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/ValueMetaData.cpp",
            # Device Detection C++
            "src/fiftyone_devicedetection_onpremise/cxx/src/ConfigDeviceDetection.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/EngineDeviceDetection.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/EvidenceDeviceDetection.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/ResultsDeviceDetection.cpp",
            # Hash C++
            "src/fiftyone_devicedetection_onpremise/cxx/src/hash/ComponentMetaDataBuilderHash.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/hash/ComponentMetaDataCollectionHash.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/hash/ConfigHash.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/hash/EngineHash.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/hash/MetaDataHash.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/hash/ProfileMetaDataBuilderHash.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/hash/ProfileMetaDataCollectionHash.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/hash/PropertyMetaDataBuilderHash.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/hash/PropertyMetaDataCollectionForComponentHash.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/hash/PropertyMetaDataCollectionForPropertyHash.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/hash/PropertyMetaDataCollectionHash.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/hash/ResultsHash.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/hash/ValueMetaDataBuilderHash.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/hash/ValueMetaDataCollectionBaseHash.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/hash/ValueMetaDataCollectionForProfileHash.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/hash/ValueMetaDataCollectionForPropertyHash.cpp",
            "src/fiftyone_devicedetection_onpremise/cxx/src/hash/ValueMetaDataCollectionHash.cpp",
            ],
        'cflags': extra_compile_args
    })

clib = ('clib', {
        'sources': [
            # Common C
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/cache.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/collection.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/component.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/data.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/dataset.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/evidence.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/exceptionsc.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/file.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/float.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/headers.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/ip.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/list.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/memory.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/overrides.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/pool.c",
			"src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/process.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/profile.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/properties.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/property.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/pseudoheader.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/resource.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/results.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/status.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/string.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/textfile.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/threading.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/tree.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/common-cxx/value.c",
            # Device Detection C
            "src/fiftyone_devicedetection_onpremise/cxx/src/dataset-dd.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/results-dd.c",
            # Hash C
            "src/fiftyone_devicedetection_onpremise/cxx/src/hash/graph.c",
            "src/fiftyone_devicedetection_onpremise/cxx/src/hash/hash.c",
        ],
        'cflags': cflags
    })

setup(
    cmdclass={"build_ext": NoSuffixBuilder},
    name="fiftyone_devicedetection_onpremise",
    version=read("version.txt"),
    author="51Degrees Engineering",
    author_email="engineering@51degrees.com",
    description = """This project contains 51Degrees Device Detection OnPremise engine that can be used with the 51Degrees Pipeline API. The Pipeline is a generic web request intelligence and data processing solution with the ability to add a range of 51Degrees and/or custom plug ins (Engines)""",
    long_description=read("readme.md"),
    long_description_content_type="text/markdown",
    python_requires=">=3.8",
    license="EUPL-1.2",
    libraries=[cpplib, clib],
    ext_package="fiftyone_devicedetection_onpremise",
    ext_modules=[DeviceDetectionHashEngineModule],
    py_modules=["fiftyone_devicedetection_onpremise"],
    packages=["fiftyone_devicedetection_onpremise"],
    package_dir={"": "src"},
    install_requires=["fiftyone_devicedetection_shared", "fiftyone_pipeline_engines_fiftyone"],
)
